unit UCube;

interface

uses
  SysUtils,
  Classes,
  Math;

const
  kCubeMask = '( %d, %d, %d )';
  kUsedCubeMask = '%d,%d,%d,%s';

type

  TCube = class(TObject)
  private
    FX: Integer;
    FY: Integer;
    FZ: Integer;
    FUsedByPiece: Integer;
  public
    constructor Create(aX: Integer = 0; aY: Integer = 0; aZ: Integer = 0);
    constructor AddCoords(c1: TCube; c2: TCube);

    function GetString: string;
    function IsUsed: boolean;

    property X: Integer read FX write FX;
    property Y: Integer read FY write FY;
    property Z: Integer read FZ write FZ;
    property UsedByPiece: integer read FUsedByPiece write FUsedByPiece;
    property Used: boolean read IsUsed;
  end;

  TCubeList = class(TList)
  private

  public
    procedure Clear; override;

    function IsInList(_Cube: TCube): integer;
    function IsInListAndNotUsed(_Cube: TCube): integer;
    function NotUsedMinCube: integer;

    function GetString: string;

    procedure SortByMin;
    procedure SortForDrawing;
    procedure MoveMinCoordTo000;
  end;

implementation

//----------------------------------------------------------------------------------------------------------------------
constructor TCube.Create(aX: Integer = 0; aY: Integer = 0; aZ: Integer = 0);
begin
  inherited Create;

  FX:= aX;
  FY:= aY;
  FZ:= aZ;
  FUsedByPiece:= -1;
end;

//----------------------------------------------------------------------------------------------------------------------
// Constructeur de cube qui ajoute les coordonnes des cubes transmis en parametre
constructor TCube.AddCoords(c1: TCube; c2: TCube);
begin
  FX:= c1.FX + c2.FX;
  FY:= c1.FY + c2.FY;
  FZ:= c1.FZ + c2.FZ;
end;

//----------------------------------------------------------------------------------------------------------------------
function TCube.IsUsed: boolean;
begin
  result:= (FUsedByPiece > -1);
end;

//----------------------------------------------------------------------------------------------------------------------
function TCube.GetString: string;
begin
  result:= Format(kCubeMask, [FX, FY, FZ]);
end;


//----------------------------------------------------------------------------------------------------------------------
//Tri choisit pour l'algorithme pour ordonner les cubes
//Tri X min, puis Y min, puis Z min
//La description choisit du minimum ne change rien  la rsolution
// tant que le puzzle et les angles de vues des pices sont tris de la meme faon
function CompareCoordinates(Item1, Item2: Pointer): Integer;
begin
  result:= -1;

  if (  TCube(Item2).X < TCube(Item1).X ) then
  begin
    result:= 1;
  end else if (TCube(Item2).X = TCube(Item1).X) then
  begin
    if ( TCube(Item2).Y < TCube(Item1).Y ) then
    begin
      result:= 1
    end else if (TCube(Item2).Y = TCube(Item1).Y) then
    begin
      if (TCube(Item2).Z < TCube(Item1).Z ) then
        result:= 1
      else if (TCube(Item2).Z = TCube(Item1).Z) then
        result:= 0;
    end;
  end;

end;

//--------------------------------------------------------------------------------------------------------------
// Tri pour le bitmap X min, puis Y max, puis Z max
// Pour le bitmap, ce tri s'explique par le fait que les cubes doivent tre trac dans
// le bon ordre pour se recouvrir de la bonne manire
// ainsi pour deux cubes de meme coordones X et Y, c'est celui qui a le Z le
// plus grand qui sera recouvert par l'autre
// cela n'est pas le cas pour Y, c'est le contraire, cependant on a un signe ">"
// car la vue est invers dans le bitmap, les Y n'tant pas dans la mme direction entre les pixels et la vue choisi
function CompareCoordinates2(Item1, Item2: Pointer): Integer;
begin
  result:= -1;

  if (  TCube(Item2).X < TCube(Item1).X ) then
  begin
    result:= 1;
  end else if (TCube(Item2).X = TCube(Item1).X) then
  begin
    if ( TCube(Item2).Y > TCube(Item1).Y ) then
    begin
      result:= 1
    end else if (TCube(Item2).Y = TCube(Item1).Y) then
    begin
      if (TCube(Item2).Z > TCube(Item1).Z ) then
        result:= 1
      else if (TCube(Item2).Z = TCube(Item1).Z) then
        result:= 0;
    end;
  end;

end;

//----------------------------------------------------------------------------------------------------------------------
//  TCubeList
//----------------------------------------------------------------------------------------------------------------------
procedure TCubeList.Clear;
var
  i: integer;
begin
  for i:= self.Count - 1 downto 0 do
    TCube(self.Items[i]).Free;

  inherited Clear;
end;

//----------------------------------------------------------------------------------------------------------------------
// Vrifie qu'un cube appartienne  la liste
function TCubeList.IsInList(_Cube: TCube): integer;
var
  i: integer;
begin
  result:= -1;

  for i:= 0 to self.Count - 1 do
    with TCube(self.Items[i]) do
      if (_Cube.X = X) then
        if (_Cube.Y = Y) then
          if (_Cube.Z = Z) then
          begin
            result:= i;
            break;
          end;
end;

//----------------------------------------------------------------------------------------------------------------------
// Vrifie qu'un cube appartienne  la liste et qu'il ne soit pas utilise
function TCubeList.IsInListAndNotUsed(_Cube: TCube): integer;
var
  i: integer;
begin
  result:= -1;

  for i:= 0 to self.Count - 1 do
    with TCube(self.Items[i]) do
      if (_Cube.X = X) then
        if (_Cube.Y = Y) then
          if (_Cube.Z = Z) then
            if (Not Used) then
            begin
              result:= i;
              break;
            end;
end;


//----------------------------------------------------------------------------------------------------------------------
// Renvoie l'indice du cube mimimum non utilise de la liste
function TCubeList.NotUsedMinCube: integer;
var
  i: integer;
begin
  result:= -1;

  for i:= 0 to Self.Count - 1 do
    if ( not TCube(Self.Items[i]).IsUsed ) then
    begin
      result:= i;
      Break;
    end;

end;

//----------------------------------------------------------------------------------------------------------------------
procedure TCubeList.SortByMin;
begin
  Self.Sort(@CompareCoordinates);
end;

//----------------------------------------------------------------------------------------------------------------------
procedure TCubeList.SortForDrawing;
begin
  Self.Sort(@CompareCoordinates2);
end;

//----------------------------------------------------------------------------------------------------------------------
// ramne le cube minimum de la liste  (0,0,0)
procedure TCubeList.MoveMinCoordTo000;
var
  i, a, b, c: integer;
begin
  //On s'assure d'abord d'ordonn les cubes
  Self.SortByMin;

  //On rcupre les coordones du cube minimum
  with TCube(Items[0]) do
  begin
    a:= X;
    b:= Y;
    c:= Z;
  end;

  //On soustrait  tous les cubes les coordones du cube minimum
  //Ramenant ainsi le cube min  (0,0,0) et conservant la cohrence de la liste
  for i:= 0 to Self.Count - 1 do
    with TCube(Items[i]) do
    begin
      X:= X - a;
      Y:= Y - b;
      Z:= Z - c;
    end;

end;

//----------------------------------------------------------------------------------------------------------------------
function TCubeList.GetString: string;
var
  i: integer;
begin
  result:= '';
  for i:= 0 to Self.Count - 1 do
    result:= result + TCube(Self.Items[i]).GetString;

end;

end.
